# Copyright (c) 2012-2018, 2021 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import errno
import datetime
import itertools
import json
import re
from collections import OrderedDict
import ipaddress

from dnstli.log import *
from dnstli.util import iso8601now
import dnstli.conf

# Get our configuration dictionary.  See dnstli.conf for the required_config_keys and optional_config_keys.
# get_config() may raise an error, which we let kill the process
conf = dnstli.conf.get_config()
    
apikey_pattern = re.compile('^[A-Z0-9+-/=]+$', re.IGNORECASE)

class HTTPResponse(Exception):
    __slots__ = ('status', 'headers', 'message',)
    def __init__(self, status, message):
        self.status = status
        self.message = message + '\n'
        self.headers = [('Content-Type', 'text/plain')]

AuthSuccess = HTTPResponse(
    '200 ok', 'Success: Authentication succeeded')
AuthNeeded = HTTPResponse(
    '403 forbidden', 'Error: Authentication needed')
AuthFail = HTTPResponse(
    '403 forbidden', 'Error: Authentication failed')

MissingAPIKey = HTTPResponse(
    '403 forbidden', 'Error: X-API-Key header not present')
BadAPIKey = HTTPResponse(
    '403 forbidden', 'Error: Bad API key')
NoRateLimits = HTTPResponse(
    '403 forbidden', 'Error: Rate Limits are not supported')

RemoteAddressUnauthorized = HTTPResponse(
    '403 forbidden', 'Client IP address not authorized for account')

InternalServerError = HTTPResponse(
    '500 internal server error', 'Error: exception occurred')

def log_login(env, username, logtype):
    msg = OrderedDict()
    msg['timestamp'] = iso8601now()
    msg['logtype'] = logtype
    msg['host'] = env['SERVER_NAME']
    msg['username'] = username
    if 'dnstli.authgroup' in env:
        msg['group'] = env['dnstli.authgroup']
    msg['ip'] = env['REMOTE_ADDR']
    log(NOTICE, json.dumps(msg))

empty_policy = object()
empty_acl = object()

# check if the REMOTE_ADDR is in our acl, if those are set.
# raise error if a problem and want_raise = True.
def check_acl(env, want_raise=True):
    if not env.get('dnstli.policy'):
        return empty_policy
    if not env['dnstli.policy'].get('addresses'):
        return empty_acl
    try:
        remote_ip = ipaddress.ip_address(unicode(env['REMOTE_ADDR']))
    except:
        # weird -- nginx/uwsgi passed us an invalid IP
        raise InternalServerError
    for acl in itertools.chain(
            conf.get('admin_addresses', []),
            env['dnstli.policy']['addresses'],
            ):
        # check if REMOTE_ADDR is equal or inside the IP or CIDR
        # block specified by acl
        if remote_ip in ipaddress.ip_network(unicode(acl)):
            return acl
    else:                       # nothing matched
        if want_raise:
            raise RemoteAddressUnauthorized
        return None

class noauth():
    def __init__(self, app):
        self.app = app

    def __call__(self, env, start_response):
        try:
            method, args = dnstli.util.get_method_args(env)

            env['dnstli.authtype'] = 'open'
            env['dnstli.authuser'] = ''
            env['dnstli.authgroup'] = None
            env['dnstli.policy'] = None

            if method == 'lookup' and args == 'auth':
                headers = [('Content-Type', 'text/plain')]
                start_response(AuthSuccess.status, headers)
                return [AuthSuccess.message]

            return self.app(env, start_response)
        except Exception:
            log_exception()
            raise InternalServerError

class apikey():
    def __init__(self, app):
        self.app = app

    def load_key(self, env):
        if not 'HTTP_X_API_KEY' in env:
            raise MissingAPIKey

        try:
            api_key = env['HTTP_X_API_KEY']
            if apikey_pattern.match(api_key):
                key_fname = os.path.join(dnstli.conf.API_DIR, api_key)
                #$$$ maybe cache based upon the last-mod-time to avoid reopening and parsing it for every query?
                keyConfDict = json.load(open(key_fname))

                env['dnstli.authtype'] = 'api'
                env['dnstli.authuser'] = keyConfDict['username']
                env['dnstli.authgroup'] = keyConfDict.get('group')
                env['dnstli.policy'] = keyConfDict.get('policy')
            else:
                raise BadAPIKey
        except IOError as e:
            if e.errno == errno.ENOENT:
                raise BadAPIKey
            else:
                log_exception()
                raise BadAPIKey
        except Exception:
            log_exception()
            raise BadAPIKey

    def __call__(self, env, start_response):
        try:
            try:
                self.load_key(env)
                check_acl(env, want_raise=True) # will raise if our IP is not authorized

                method, args = dnstli.util.get_method_args(env)
                if method == 'lookup' and args == 'rate_limit':
                    raise NoRateLimits
                return self.app(env, start_response)
            except HTTPResponse, res:
                start_response(res.status, res.headers)
                return [res.message]
        except Exception:
            log_exception()
            raise InternalServerError
